import yul.yul_ast as ast
from yul.AstMapper import AstMapper

CAIRO_KEYWORDS = {"ret", "felt", "jmp", "func", "end"}


def mangle(identifier: str) -> str:
    identifier = identifier.replace("$", "_")
    if identifier in CAIRO_KEYWORDS:
        return identifier + "__warp_mangled"
    else:
        return identifier


class MangleNamesVisitor(AstMapper):
    def visit_typed_name(self, node: ast.TypedName):
        return ast.TypedName(mangle(node.name), node.type)

    def visit_identifier(self, node: ast.Identifier):
        return ast.Identifier(name=mangle(node.name))

    def visit_function_definition(self, node: ast.FunctionDefinition):
        return ast.FunctionDefinition(
            name=mangle(node.name),
            parameters=[self.visit(x) for x in node.parameters],
            return_variables=[self.visit(x) for x in node.return_variables],
            body=self.visit(node.body),
        )



from typing import List, Optional

from eth_utils import function_abi_to_4byte_selector
from web3 import Web3
from yul.utils import cairoize_bytes


def get_evm_calldata(
    abi: dict, abi_original: dict, bytecode, fn_name: str, inputs: List[int]
) -> str:
    w3 = Web3()
    contract_abi_original = w3.eth.contract(abi=abi_original, bytecode=bytecode)
    contract = w3.eth.contract(abi=abi, bytecode=bytecode)
    for f in contract_abi_original.all_functions():
        if f.function_identifier == fn_name:
            selector = function_abi_to_4byte_selector(f.abi).hex()
    calldata = "0x" + selector + contract.encodeABI(fn_name=fn_name, args=inputs)[10:]
    return calldata


def evm_to_cairo_calldata(
    abi: dict,
    abi_original: dict,
    bytecode: str,
    fn_name: str,
    inputs: List[int],
    address: Optional[int] = None,
):
    evm_calldata = get_evm_calldata(
        abi=abi,
        abi_original=abi_original,
        bytecode=bytecode,
        fn_name=fn_name,
        inputs=inputs,
    )
    cairo_input, unused_bytes = cairoize_bytes(bytes.fromhex(evm_calldata[2:]))
    calldata_size = (len(cairo_input) * 16) - unused_bytes
    if address is None:
        return [calldata_size, len(cairo_input), *cairo_input]
    else:
        return [calldata_size, len(cairo_input), *cairo_input, address]



import json
import os
import subprocess
import sys
from http import HTTPStatus
from typing import Any, Dict, Optional, Union

import aiohttp
from cli.StarkNetEvmContract import evm_to_cairo_calldata, get_evm_calldata
from eth_hash.auto import keccak
from starkware.starknet.definitions import fields
from starkware.starknet.services.api.contract_definition import ContractDefinition
from starkware.starknet.services.api.gateway.transaction import (
    Deploy,
    InvokeFunction,
    Transaction,
)
from yul.utils import cairoize_bytes

WARP_ROOT = os.path.abspath(os.path.join(__file__, "../.."))


def get_selector_cairo(args: str) -> int:
    return int.from_bytes(keccak(args.encode("ascii")), "big") & (2 ** 250 - 1)


async def send_req(method, url, tx: Optional[Union[str, Dict[str, Any]]] = None):
    if tx is not None:
        async with aiohttp.ClientSession() as session:
            async with session.request(
                method=method, url=url, data=Transaction.Schema().dumps(obj=tx)
            ) as response:
                text = await response.text()
                return text
    else:
        async with aiohttp.ClientSession() as session:
            async with session.request(method=method, url=url, data=None) as response:
                text = await response.text()
                return text


# returns true/false on transaction success/failure
async def _invoke(program_info: dict, address, function, evm_inputs):
    try:
        address = int(address, 16)
    except ValueError as e:
        raise ValueError("Invalid address format.") from e
    abi = program_info["sol_abi"]
    bytecode = program_info["sol_bytecode"]
    calldata = evm_to_cairo_calldata(abi, bytecode, function, evm_inputs, address)
    selector = get_selector_cairo("fun_ENTRY_POINT")
    tx = InvokeFunction(
        contract_address=address,
        entry_point_selector=selector,
        calldata=calldata,
        signature=[],
    )

    response = await send_req(
        method="POST", url="https://alpha3.starknet.io/gateway/add_transaction", tx=tx
    )
    tx_hash = json.loads(response)["transaction_hash"]
    print(
        f"""\
Invoke transaction was sent.
Contract address: 0x{address:064x}.
Transaction hash: {tx_hash}."""
    )
    return True


def starknet_compile(cairo_path, contract_base):
    compiled = f"{contract_base}_compiled.json"
    process = subprocess.Popen(
        [
            "starknet-compile",
            "--disable_hint_validation",
            f"{cairo_path}",
            "--output",
            compiled,
            "--cairo_path",
            f"{WARP_ROOT}/cairo-src",
        ]
    )
    output = process.wait()
    if output == 1:
        raise Exception("Compilation failed")
    return compiled


async def _deploy(cairo_path, contract_base, program_info, constructor_args):
    compiled_contract = starknet_compile(cairo_path, contract_base)
    address = fields.ContractAddressField.get_random_value()
    with open(compiled_contract) as f:
        cont = f.read()

    contract_definition = ContractDefinition.loads(cont)
    tx = get_constructor_tx(
        contract_definition=contract_definition,
        program_info=program_info,
        address=address,
        constructor_args=constructor_args,
    )
    url = "https://alpha3.starknet.io/gateway/add_transaction"

    async with aiohttp.ClientSession() as session:
        async with session.request(
            method="POST", url=url, data=Transaction.Schema().dumps(obj=tx)
        ) as response:
            text = await response.text()
            if response.status != HTTPStatus.OK:
                sys.exit(f"FAIL:\n{response}")
            tx_hash = json.loads(text)["transaction_hash"]
    address_path = f"{os.path.abspath(contract_base)}_ADDRESS.txt"
    with open(address_path, "w") as f:
        f.write(f"0x{address:064x}")
    print(
        f"""\
Deploy transaction was sent.
Contract address: 0x{address:064x}.
Transaction hash: {tx_hash}.

Contract Address Has Been Written to {address_path}
"""
    )


def get_constructor_tx(
    contract_definition, program_info: dict, address: int, constructor_args: list
) -> Deploy:
    if constructor_args == "\0":
        tx = Deploy(
            contract_address_salt=fields.ContractAddressSalt.get_random_value(),
            contract_definition=contract_definition,
            constructor_calldata=[],
        )
    elif "constructor" in program_info["dynamic_argument_functions"]:
        calldata = evm_to_cairo_calldata(
            program_info["abi"],
            program_info["bytecode"],
            "__warp_ctorHelper_DynArgs",
            constructor_args,
        )
        tx = Deploy(
            contract_address_salt=fields.ContractAddressSalt.get_random_value(),
            contract_definition=contract_definition,
            constructor_calldata=calldata,
        )
    else:
        flattened_args = list(flatten(constructor_args))
        split_args = []
        for arg in flattened_args:
            high, low = divmod(arg, 2 ** 128)
            split_args += [low, high]
        tx = Deploy(
            contract_address_salt=address,
            contract_definition=contract_definition,
            constructor_calldata=split_args,
        )
    return tx


def flatten(l):
    for i in l:
        if isinstance(i, int):
            yield i
        else:
            yield from flatten(i)


async def _status(tx_hash):
    status = f"https://alpha3.starknet.io/feeder_gateway/get_transaction_status?transactionHash={tx_hash}"
    res = await send_req("GET", status)
    print(json.loads(res))



from __future__ import annotations

import re
from collections import defaultdict
from contextlib import contextmanager
from typing import Callable, Optional

import yul.yul_ast as ast
from yul.AstVisitor import AstVisitor
from yul.BuiltinHandler import BuiltinHandler
from yul.FunctionGenerator import CairoFunctions
from yul.implicits import IMPLICITS_SET, copy_implicit, print_implicit
from yul.Imports import format_imports, merge_imports
from yul.NameGenerator import NameGenerator
from yul.storage_access import (
    StorageVar,
    extract_var_from_getter,
    extract_var_from_setter,
    generate_getter_body,
    generate_setter_body,
    generate_storage_var_declaration,
)
from yul.utils import HANDLERS_DECL

UINT128_BOUND = 2 ** 128

COMMON_IMPORTS = {
    "starkware.cairo.common.registers": {"get_fp_and_pc"},
    "starkware.cairo.common.dict_access": {"DictAccess"},
    "starkware.cairo.common.math_cmp": {"is_le"},
    "starkware.cairo.common.default_dict": {
        "default_dict_new",
        "default_dict_finalize",
    },
    "starkware.cairo.common.uint256": {"Uint256", "uint256_eq"},
    "starkware.cairo.common.math": {"unsigned_div_rem"},
    "starkware.cairo.common.cairo_builtins": {
        "HashBuiltin",
        "BitwiseBuiltin",
    },
    "starkware.cairo.common.alloc": {"alloc"},
    "evm.exec_env": {"ExecutionEnvironment"},
    "evm.utils": {"update_msize"},
    "evm.array": {"array_create_from_memory", "array_copy_to_memory"},
}

MAIN_PREAMBLE = """%lang starknet
%builtins pedersen range_check bitwise
"""


class ToCairoVisitor(AstVisitor):
    def __init__(
        self,
        public_functions: list[str],
        function_mutabilities: dict[str, str],
        name_gen: NameGenerator,
        cairo_functions: CairoFunctions,
        builtins_map: Callable[[CairoFunctions], dict[str, BuiltinHandler]],
    ):
        super().__init__()
        self.public_functions = public_functions
        self.function_mutabilities = function_mutabilities
        self.name_gen = name_gen
        self.cairo_functions = cairo_functions
        self.builtins_map = builtins_map(self.cairo_functions)
        self.dynamic_argument_functions: list[str] = []
        self.imports = defaultdict(set)
        merge_imports(self.imports, COMMON_IMPORTS)
        self.last_function: Optional[ast.FunctionDefinition] = None
        self.last_used_implicits: tuple[str, ...] = ()
        self.function_to_implicits: dict[str, set[str]] = {}
        self.in_entry_function: bool = False
        self.storage_variables: set[StorageVar] = set()

    #         # We need these functions to be generated by default
    #         self.cairo_functions.bigstruct(128)
    #         self.cairo_functions.bigstructencode_function(128)
    #         self.cairo_functions.bigstructdecode_function(128)

    def translate(self, node: ast.Node) -> str:
        main_part = self.print(node)
        storage_vars = self.storage_variables | self.cairo_functions.storage_vars
        storage_var_decls = [
            generate_storage_var_declaration(x) for x in sorted(storage_vars)
        ]
        return "\n".join(
            [
                MAIN_PREAMBLE,
                format_imports(self.imports),
                "",
                *self.cairo_functions.get_definitions(),
                "",
                *storage_var_decls,
                HANDLERS_DECL,
                main_part,
            ]
        )

    def print(self, node: ast.Node, *args, **kwargs) -> str:
        return self.visit(node, *args, **kwargs)

    def common_visit(self, node: ast.Node, *args, **kwargs):
        raise AssertionError(
            f"Each node type should have a custom visit, but {type(node)} doesn't"
        )

    def visit_typed_name(self, node: ast.TypedName) -> str:
        return f"{node.name} : {node.type}"

    def visit_literal(self, node: ast.Literal) -> str:
        v = int(node.value)  # to convert bools: True -> 1, False -> 0
        high, low = divmod(v, UINT128_BOUND)
        return f"Uint256(low={low}, high={high})"

    def visit_identifier(self, node: ast.Identifier) -> str:
        return f"{node.name}"

    def visit_assignment(self, node: ast.Assignment) -> str:
        return self.visit(
            ast.VariableDeclaration(
                variables=[ast.TypedName(x.name) for x in node.variable_names],
                value=node.value,
            )
        )

    def visit_function_call(self, node: ast.FunctionCall) -> str:
        fun_repr = self.print(node.function_name)
        args = [self.print(x) for x in node.arguments]
        if fun_repr == "revert" and not self.in_entry_function:
            return "assert 0 = 1\njmp rel 0"
        if fun_repr == "revert" and self.in_entry_function:
            return ""
        if fun_repr == "pop":
            return ""
        result: str
        builtin_handler = self.builtins_map.get(fun_repr)
        if builtin_handler:
            merge_imports(self.imports, builtin_handler.required_imports())
            self.last_used_implicits = builtin_handler.get_used_implicits()
            if self.in_entry_function and "mstore" in fun_repr:
                result = (
                    f"with memory_dict, msize, range_check_ptr:\n"
                    f"{builtin_handler.get_function_call(args)}\n"
                    f"end\n"
                )
            else:
                result = f"{builtin_handler.get_function_call(args)}"
        else:
            self.last_used_implicits = sorted(
                self.function_to_implicits.setdefault(
                    node.function_name.name, IMPLICITS_SET
                )
            )
            args_repr = ", ".join(self.print(x) for x in node.arguments)
            if self.in_entry_function and "__warp_if_" in fun_repr:
                result = (
                    f"with exec_env, memory_dict, msize, pedersen_ptr, range_check_ptr, syscall_ptr:\n"
                    f"{fun_repr}({args_repr})\n"
                    f"end\n"
                )
            else:
                result = f"{fun_repr}({args_repr})"
        if self.last_function:
            self.function_to_implicits.setdefault(
                self.last_function.name, set()
            ).update(self.last_used_implicits)
        return result

    def visit_expression_statement(self, node: ast.ExpressionStatement) -> str:
        if isinstance(node.expression, ast.FunctionCall):
            return self.visit(node.expression)
        else:
            # see ast.ExpressionStatement docstring
            raise ValueError("What am I going to do with it? Why is it here?..")

    def visit_variable_declaration(self, node: ast.VariableDeclaration) -> str:
        if node.value is None:
            decls_repr = "\n".join(
                self.print(ast.VariableDeclaration(variables=[x], value=ast.Literal(0)))
                for x in node.variables
            )
            return decls_repr
        value_repr = self.visit(node.value)
        if not node.variables:
            return value_repr
        vars_repr = ", ".join(f"local {self.visit(x)}" for x in node.variables)
        if isinstance(node.value, ast.FunctionCall):
            return f"let ({vars_repr}) = {value_repr}"
        else:
            assert len(node.variables) == 1
            return f"{vars_repr} = {value_repr}"

    def visit_block(self, node: ast.Block) -> str:
        stmt_reprs = []
        for stmt in node.statements:
            with self._new_statement():
                stmt_reprs.append(self.visit(stmt))
                for implicit in self.last_used_implicits:
                    stmt_reprs.append(copy_implicit(implicit))
        return "\n".join(stmt_reprs)

    def visit_function_definition(self, node: ast.FunctionDefinition):
        if "ENTRY_POINT" in node.name:
            self.in_entry_function = True
        self.last_function = node
        params_repr = ", ".join(self.print(x) for x in node.parameters)
        returns_repr = ", ".join(self.print(x) for x in node.return_variables)
        body_repr = self._try_make_storage_accessor_body(node)
        if not body_repr:
            body_repr = self.print(node.body)
        if "_DynArgs" in node.name:
            if node.name == "fun_warp_constructor_DynArgs":
                self.dynamic_argument_functions.append("constructor")
            else:
                self.dynamic_argument_functions.append(node.name)
        if node.name == "fun_warp_constructor":
            return (
                f"@constructor\n"
                f"func constructor{{pedersen_ptr : HashBuiltin*, range_check_ptr,"
                f"syscall_ptr : felt* , bitwise_ptr : BitwiseBuiltin*}}({params_repr}):\n"
                f"alloc_locals\n"
                f"let (local memory_dict) = default_dict_new(0)\n"
                f"local memory_dict_start: DictAccess* = memory_dict\n"
                f"let msize = 0\n"
                f"with pedersen_ptr, range_check_ptr, bitwise_ptr, memory_dict, msize:\n"
                f"{body_repr}\n"
                f"end\n"
                f"end"
            )
        elif node.name == "fun_warp_constructor_DynArgs":
            return (
                f"@constructor\n"
                f"func constructor{{pedersen_ptr : HashBuiltin*, range_check_ptr,"
                f"syscall_ptr : felt* , bitwise_ptr : BitwiseBuiltin*}}(calldata_size,"
                f"calldata_len, calldata : felt*):\n"
                f"alloc_locals\n"
                f"local pedersen_ptr : HashBuiltin* = pedersen_ptr\n"
                f"local range_check_ptr = range_check_ptr\n"
                f"local syscall_ptr : felt* = syscall_ptr\n"
                f"let (returndata_ptr: felt*) = alloc()\n"
                f"local exec_env : ExecutionEnvironment = ExecutionEnvironment("
                f"calldata_size=calldata_size, calldata_len=calldata_len, calldata=calldata,"
                f"returndata_size=0, returndata_len=0, returndata=returndata_ptr,"
                f"to_returndata_size=0, to_returndata_len=0, to_returndata=returndata_ptr)\n"
                f"let (local memory_dict) = default_dict_new(0)\n"
                f"local memory_dict_start: DictAccess* = memory_dict\n"
                f"let msize = 0\n"
                f"with pedersen_ptr, range_check_ptr, bitwise_ptr, memory_dict, msize, exec_env:\n"
                f"{body_repr}\n"
                f"end\n"
                f"end\n"
            )

        if node.name == "fun_ENTRY_POINT":
            # The leave gets replaced with the wrong return type in this case
            # we need to replace it with our return
            body_repr = re.sub("return \(\)$", "", body_repr)
            returns_repr = "success: felt, returndata_size: felt, returndata_len: felt, returndata: felt*"
            self.in_entry_function = False
            return (
                "@external\n"
                f"func {node.name}{{pedersen_ptr : HashBuiltin*, range_check_ptr,"
                f"syscall_ptr : felt* , bitwise_ptr : BitwiseBuiltin*}}(calldata_size,"
                f"calldata_len, calldata : felt*, self_address : felt) -> ({returns_repr}):\n"
                f"alloc_locals\n"
                f"initialize_address{{range_check_ptr=range_check_ptr, syscall_ptr=syscall_ptr, pedersen_ptr=pedersen_ptr}}(self_address)\n"
                f"local pedersen_ptr : HashBuiltin* = pedersen_ptr\n"
                f"local range_check_ptr = range_check_ptr\n"
                f"local syscall_ptr : felt* = syscall_ptr\n"
                f"let (returndata_ptr: felt*) = alloc()\n"
                f"local exec_env : ExecutionEnvironment = ExecutionEnvironment("
                f"calldata_size=calldata_size, calldata_len=calldata_len, calldata=calldata,"
                f"returndata_size=0, returndata_len=0, returndata=returndata_ptr,"
                f"to_returndata_size=0, to_returndata_len=0, to_returndata=returndata_ptr)\n"
                f"let (local memory_dict) = default_dict_new(0)\n"
                f"local memory_dict_start: DictAccess* = memory_dict\n"
                f"let msize = 0\n"
                f"with exec_env, msize, memory_dict:\n"
                f"  {body_repr}\n"
                f"end\n"
                f"default_dict_finalize(memory_dict_start, memory_dict, 0)\n"
                f"return (1, exec_env.to_returndata_size, exec_env.to_returndata_len, exec_env.to_returndata)\n"
                f"end\n"
            )

        function_name = extract_function_name(node.name) or node.name
        mutability = self.function_mutabilities.get(function_name)
        mutability = "view" if mutability in ["pure", "view"] else "external"
        if node.name.startswith("getter"):
            mutability = "view"
        elif node.name.startswith("setter"):
            mutability = "external"

        external = node.name in self.public_functions
        external_function = (
            self._make_external_function(node, mutability) if external else ""
        )
        if "warp_call" in body_repr or "warp_static_call" in body_repr:
            self.function_to_implicits[node.name].add("bitwise_ptr")
        implicits = sorted(self.function_to_implicits.setdefault(node.name, set()))
        implicits_decl = ""
        if implicits:
            implicits_decl = ", ".join(print_implicit(x) for x in implicits)
            if (
                "exec_env" in self.function_to_implicits[node.name]
                and "exec_env" not in implicits
                and "block_00" not in node.name
            ):
                implicits_decl = "{exec_env : ExecutionEnvironment, " + implicits_decl
            else:
                implicits_decl = "{" + implicits_decl
            if not "range_check" in implicits_decl:
                implicits_decl += (
                    ", range_check_ptr}"
                    if len(implicits_decl) > 1
                    and implicits_decl != "{exec_env : ExecutionEnvironment, "
                    else "range_check_ptr}"
                )
            else:
                implicits_decl += "}"
        # We do not want to annotate current function with mutability if we are going to
        # generate a corresponding external function
        mutability = (
            f"@{mutability}"
            if not external and self.function_mutabilities.get(function_name)
            else ""
        )
        self.in_entry_function = False
        return (
            f"{mutability}\n"
            f"func {node.name}{implicits_decl}({params_repr}) -> ({returns_repr}):\n"
            f"alloc_locals\n"
            f"{body_repr}\n"
            f"end\n"
            f"{external_function}"
        )

    def visit_if(self, node: ast.If) -> str:
        cond_repr = self.print(node.condition)
        body_repr = self.print(node.body)
        else_repr = ""
        if node.else_body:
            else_repr = f"else:\n\t{self.print(node.else_body)}\n"
        return (
            f"if {cond_repr}.low + {cond_repr}.high != 0:\n"
            f"\t{body_repr}\n"
            f"{else_repr}"
            f"end"
        )

    def visit_case(self, node: ast.Case):
        return AssertionError("There should be no cases, run SwitchToIfVisitor first")

    def visit_switch(self, node: ast.Switch):
        return AssertionError(
            "There should be no switches, run SwitchToIfVisitor first"
        )

    def visit_for_loop(self, node: ast.ForLoop):
        raise AssertionError(
            "There should be no for loops, run ForLoopEliminator first"
        )

    def visit_break(self, node: ast.Break):
        raise AssertionError("There should be no breaks, run ForLoopEliminator first")

    def visit_continue(self, node: ast.Continue):
        raise AssertionError(
            "There should be no continues, run ForLoopEliminator first"
        )

    def visit_leave(self, _node: ast.Leave) -> str:
        return_names = ", ".join(x.name for x in self.last_function.return_variables)
        return f"return ({return_names})"

    def _make_external_function(
        self, node: ast.FunctionDefinition, mutability: str
    ) -> str:
        params = ", ".join(self.print(x) for x in node.parameters)
        returns = ", ".join(self.print(x) for x in node.return_variables)
        inner_args = ", ".join(x.name for x in node.parameters)
        inner_returns = ", ".join(
            f"local {self.visit(x)}" for x in node.return_variables
        )
        inner_call = f"{node.name}({inner_args})"
        inner_assignment = (
            f"let ({inner_returns}) = {inner_call}"
            if node.return_variables
            else inner_call
        )
        external_returns = ", ".join(
            f"{x.name}={x.name}" for x in node.return_variables
        )
        implicits = sorted(IMPLICITS_SET - {"msize", "memory_dict", "exec_env"})
        implicits_repr = ", ".join(print_implicit(x) for x in implicits)
        implicit_copy = "\n".join(copy_implicit(x) for x in implicits)

        inner_implicits = "memory_dict, msize"
        init_exec_env = ""
        if "exec_env" in self.function_to_implicits[node.name]:
            params += ", " if params else ""
            params += "calldata_size, calldata_len, calldata : felt*"
            init_exec_env = (
                f"let (returndata_ptr : felt*) = alloc()\n"
                f"local exec_env : ExecutionEnvironment ="
                f"ExecutionEnvironment(calldata_size=calldata_size,"
                f"calldata_len=calldata_len, calldata=calldata, returndata_size=0, returndata_len=0, returndata=returndata_ptr,"
                f"to_returndata_size=0, to_returndata_len=0, to_returndata=returndata_ptr)\n"
            )
            inner_implicits += ", exec_env"
        return (
            f"\n@{mutability}\n"
            f"func {node.name}_external"
            f"{{{implicits_repr}}}"
            f"({params}) -> ({returns}):\n"
            f"alloc_locals\n"
            f"let (local memory_dict) = default_dict_new(0)\n"
            f"local memory_dict_start : DictAccess* = memory_dict\n"
            f"let msize = 0\n"
            f"{init_exec_env}"
            f"with {inner_implicits}:\n"
            f"  {inner_assignment}\n"
            f"end\n"
            f"{implicit_copy}\n"
            f"default_dict_finalize(memory_dict_start, memory_dict, 0)\n"
            f"return ({external_returns})\n"
            f"end\n\n"
        )

    @contextmanager
    def _new_statement(self):
        old = self.last_used_implicits
        self.last_used_implicits = ()
        try:
            yield None
        finally:
            self.last_used_implicits = old

    def _try_make_storage_accessor_body(
        self, node: ast.FunctionDefinition
    ) -> Optional[str]:
        getter_var = extract_var_from_getter(node.name)
        setter_var = extract_var_from_setter(node.name)
        if not (getter_var or setter_var):
            return None

        self.function_to_implicits.setdefault(node.name, set()).update(
            ("pedersen_ptr", "range_check_ptr", "syscall_ptr")
        )
        accessor_args = tuple(x.name for x in node.parameters)
        if getter_var:
            assert (
                len(node.return_variables) == 1
            ), "We don't support multivalued storage variables yet"
            name = getter_var
            arg_types = tuple(x.type for x in node.parameters)
            res_type = node.return_variables[0].type
            body = generate_getter_body(getter_var, accessor_args)
        else:
            assert setter_var
            assert (
                node.parameters
            ), "Storage variable setters must have at least one parameter (the value to set)"
            name = setter_var
            arg_types = tuple(x.type for x in node.parameters[:-1])
            res_type = node.parameters[-1].type
            body = generate_setter_body(setter_var, accessor_args)
        self.storage_variables.add(
            StorageVar(name=name, arg_types=arg_types, res_type=res_type)
        )
        return body


def extract_function_name(name: str) -> Optional[str]:
    if name.startswith("getter"):
        return extract_var_from_getter(name)
    elif name.startswith("setter"):
        return extract_var_from_setter(name)
    else:
        fun_pattern = re.compile(r"fun_(\S*)(_(\d+))?")
        match = re.fullmatch(fun_pattern, name)
        if not match:
            return None
        return match.group(1)



"""This module contains definitions of Yul AST nodes, mostly following
the specification here:
https://docs.soliditylang.org/en/v0.8.7/yul.html#specification-of-yul.

"""

from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
from functools import lru_cache
from typing import Iterable, Optional, Union

from yul.utils import snakify


class Node(ABC):
    pass


@dataclass(eq=False, frozen=True)
class TypedName(Node):
    name: str
    type: str = "Uint256"


@dataclass(eq=True, frozen=True)
class Literal(Node):
    value: Union[int, bool]


@dataclass(eq=True, order=True, frozen=True)
class Identifier(Node):
    name: str


@dataclass(eq=False, frozen=True)
class Assignment(Node):
    variable_names: list[Identifier]
    value: "Expression"


@dataclass(eq=False, frozen=True)
class FunctionCall(Node):
    function_name: Identifier
    arguments: list["Expression"]


@dataclass(eq=False, frozen=True)
class ExpressionStatement(Node):
    """According to
    https://docs.soliditylang.org/en/latest/yul.html#restrictions-on-the-grammar,
    only top-level expressions can be statements. Furthermore, they
    must evaluate to zero values. That only leaves function calls
    that return zero values. Weird.

    """

    expression: "Expression"


@dataclass(eq=False, frozen=True)
class VariableDeclaration(Node):
    variables: list[TypedName]
    value: Optional["Expression"]  # None means all variables initialize to 0


@dataclass(eq=False, frozen=True)
class Block(Node):
    statements: tuple["Statement", ...] = ()


@dataclass(eq=False, frozen=True)
class FunctionDefinition(Node):
    name: str
    parameters: list[TypedName]
    return_variables: list[TypedName]
    body: Block


@dataclass(eq=False, frozen=True)
class If(Node):
    condition: "Expression"
    body: Block
    # ↓ doesn't exist in Yul, convenient for mapping to cairo
    else_body: Optional[Block] = None


@dataclass(eq=False, frozen=True)
class Case(Node):
    value: Optional[Literal]  # None for the default case
    body: Block


@dataclass(eq=False, frozen=True)
class Switch(Node):
    expression: "Expression"
    cases: list[Case]


@dataclass(eq=False, frozen=True)
class ForLoop(Node):
    pre: Block
    condition: "Expression"
    post: Block
    body: Block


@dataclass(eq=False, frozen=True)
class Break(Node):
    pass


@dataclass(eq=False, frozen=True)
class Continue(Node):
    pass


@dataclass(eq=False, frozen=True)
class Leave(Node):
    pass


# No two nodes of this class should be different from each
# other. Thus, it's cheaper to create one object and use it in all
# contexts, rather than create new `Leave()` each time.
LEAVE: Leave = Leave()

LEAVE_BLOCK: Block = Block(statements=((LEAVE,)))


def yul_log_not(argument: Expression) -> FunctionCall:
    return FunctionCall(Identifier("iszero"), [argument])


Expression = Union[Literal, Identifier, FunctionCall]
Statement = Union[
    ExpressionStatement,
    Assignment,
    VariableDeclaration,
    FunctionDefinition,
    If,
    Switch,
    ForLoop,
    Break,
    Continue,
    Leave,
    Block,
]

NODE_TYPES = frozenset(
    (
        Literal,
        Identifier,
        FunctionCall,
        ExpressionStatement,
        Assignment,
        VariableDeclaration,
        FunctionDefinition,
        If,
        Switch,
        ForLoop,
        Break,
        Continue,
        Leave,
        Block,
    )
)



import asyncio
import json
import os
import shutil
import sysconfig
from ast import literal_eval
from enum import Enum
from tempfile import NamedTemporaryFile
import pkg_resources

import click
from cli.commands import _deploy, _invoke, _status
from yul.main import transpile_from_solidity
from yul.utils import get_low_high


class Command(Enum):
    INVOKE = 0
    CALL = 1
    DEPLOY = 2
    STATUS = 3


@click.group()
def warp():
    pass


@warp.command()
@click.argument("file_path", type=click.Path(exists=True))
@click.argument("contract_name")
def transpile(file_path, contract_name):
    path = click.format_filename(file_path)
    output = transpile_from_solidity(file_path, contract_name)
    with open(f"{file_path[:-4]}.json", "w") as f:
        json.dump(output, f)
    click.echo(f"The transpilation output has been written to {path[:-4]}.json")


@warp.command()
@click.option(
    "--program",
    required=True,
    type=click.Path(exists=True, dir_okay=False),
    help="the path to the transpiled program JSON file",
)
@click.option("--address", required=True, help="contract address")
@click.option(
    "--function",
    required=True,
    help="the name of the function to invoke, as defined in the Solidity contract",
)
@click.option("--inputs", required=True, help="Function Arguments")
def invoke(program, address, function, inputs):
    inputs = literal_eval(inputs)
    click.echo(inputs)
    click.echo(function)
    with open(program, "r") as f:

        program_info = json.load(f)
    asyncio.run(_invoke(program_info, address, function, inputs))


@warp.command()
@click.argument(
    "program",
    nargs=1,
    required=True,
    type=click.Path(exists=True, dir_okay=False),
)
@click.option("--constructor_args", required=False, default="\0")
def deploy(program, constructor_args):
    """Deploy PROGRAM.

    PROGRAM is the path to the transpiled program JSON file.

    """
    try:
        constructor_args = literal_eval(constructor_args)
    except ValueError:
        pass
    assert program.endswith(".json")
    contract_base = program[: -len(".json")]
    with open(program, "r") as pf:
        program_info = json.load(pf)
    tmp = NamedTemporaryFile(mode="w", delete=False)
    cairo_path = tmp.name
    tmp.write(program_info["cairo_code"])
    tmp.close()
    try:
        asyncio.run(_deploy(cairo_path, contract_base, program_info, constructor_args))
    finally:
        os.remove(tmp.name)


@warp.command()
@click.argument("tx_id", nargs=1, required=True)
def status(tx_id):
    asyncio.run(_status(tx_id))


def main():
    try:
        base_env_dir = os.environ['VIRTUAL_ENV']
        new_kudu_exe = os.path.join(base_env_dir,"bin/kudu")
    except KeyError:
        base_env_dir = sysconfig.get_path('scripts', f'{os.name}_user')
        new_kudu_exe = os.path.join(base_env_dir,"kudu")
    kudu_pkg_dir = os.path.join(pkg_resources.get_distribution("warp==0.2.1").location, "bin/kudu")
    if os.path.exists(new_kudu_exe):
        os.remove(new_kudu_exe)
    shutil.copy2(kudu_pkg_dir, new_kudu_exe)
    warp()



from __future__ import annotations

from typing import Dict, Set

Imports = Dict[str, Set[str]]


def format_names(import_names: set[str]) -> str:
    """Create an import names list, e.g. '(name1, name2)' in 'from module
    import (name1, name2)'.

    """
    if len(import_names) == 0:
        raise ValueError("Import names list can't be empty")
    if len(import_names) == 1:
        return next(iter(import_names))

    sorted_names = sorted(import_names)
    import_line_fmt = "    {0},"
    import_lines = [import_line_fmt.format(name) for name in sorted_names]
    return "(\n" + "\n".join(import_lines) + "\n)"


def format_imports(imports: Imports) -> str:
    sorted_imports = sorted(imports.items())
    return "\n".join(
        f"from {module} import {format_names(names)}"
        for module, names in sorted_imports
    )


def merge_imports(imports: Imports, new_imports: Imports):
    for module, names in new_imports.items():
        imports[module].update(names)



IMPLICITS = {
    "memory_dict": "DictAccess*",
    "msize": None,
    "pedersen_ptr": "HashBuiltin*",
    "range_check_ptr": None,
    "syscall_ptr": "felt*",
    "exec_env": "ExecutionEnvironment",
    "bitwise_ptr": "BitwiseBuiltin*",
}

IMPLICITS_SET = set(IMPLICITS.keys())


def print_implicit(name):
    type_ = IMPLICITS.get(name, None)
    if type_ is None:
        return name
    else:
        return f"{name}: {type_}"


def copy_implicit(name):
    return f"local {print_implicit(name)} = {name}"



from __future__ import annotations

import yul.yul_ast as ast
from yul.call_graph import build_callgraph


def top_sort_ast(block: ast.Block) -> ast.Block:
    callgraph = build_callgraph(block)
    fun_stmts = [
        stmt for stmt in block.statements if isinstance(stmt, ast.FunctionDefinition)
    ]
    other_stmts = [
        stmt
        for stmt in block.statements
        if not isinstance(stmt, ast.FunctionDefinition)
    ]
    # ↓ using dict instead of set to preserve order
    unmarked: dict[ast.FunctionDefinition, bool] = dict.fromkeys(fun_stmts)
    ordered: list[ast.FunctionDefinition] = []

    def dfs(node: ast.FunctionDefinition):
        if node not in unmarked:
            return
        del unmarked[node]
        for child in callgraph.get(node, tuple()):
            dfs(child)
        ordered.append(node)

    while unmarked:
        node = next(iter(unmarked))
        dfs(node)

    return ast.Block((*ordered, *other_stmts))



from __future__ import annotations

import yul.yul_ast as ast
from yul.AstMapper import AstMapper


class LeaveNormalizer(AstMapper):
    """This class puts a 'leave' statement at the end of each function. If
    the function consists of a single 'if', it puts 'leave' at the end of
    each branch.

    After this transformation, the only thing required to correctly
    place cairo 'return's during the transpilation process is to just
    replace yul 'leave's.

    """

    def visit_function_definition(self, node: ast.FunctionDefinition):
        return ast.FunctionDefinition(
            name=node.name,
            parameters=node.parameters,
            return_variables=node.return_variables,
            body=normalize_block(node.body),
        )


def normalize_block(block: ast.Block) -> ast.Block:
    return ast.Block(normalize_statements(block.statements))


def normalize_statements(statements: tuple[ast.Statement]) -> tuple[ast.Statement]:
    statements = strip_trailing_leaves(statements)
    if statements and isinstance(statements[-1], ast.If):
        return (*statements[:-1], normalize_if(statements[-1]))
    else:
        return (*statements, ast.LEAVE)


def normalize_if(if_node: ast.If) -> ast.If:
    return ast.If(
        condition=if_node.condition,
        body=normalize_block(if_node.body),
        else_body=normalize_block(if_node.else_body)
        if if_node.else_body
        else ast.LEAVE_BLOCK,
    )


def strip_trailing_leaves(statements: tuple[ast.Statement]) -> tuple[ast.Statement]:
    leave_no = len(statements)
    while leave_no > 0 and isinstance(statements[leave_no - 1], ast.Leave):
        leave_no -= 1
    return statements[:leave_no]



from __future__ import annotations

import json
import os
import shutil
import subprocess
import sys

import click
import yul.yul_ast as ast
from starkware.cairo.lang.compiler.parser import parse_file
from yul.BuiltinHandler import get_default_builtins
from yul.ExpressionSplitter import ExpressionSplitter
from yul.ForLoopEliminator import ForLoopEliminator
from yul.ForLoopSimplifier import ForLoopSimplifier
from yul.FunctionGenerator import CairoFunctions, FunctionGenerator
from yul.FunctionPruner import FunctionPruner
from yul.LeaveNormalizer import LeaveNormalizer
from yul.MangleNamesVisitor import MangleNamesVisitor
from yul.NameGenerator import NameGenerator
from yul.parse import parse_node
from yul.RevertNormalizer import RevertNormalizer
from yul.ScopeFlattener import ScopeFlattener
from yul.SwitchToIfVisitor import SwitchToIfVisitor
from yul.ToCairoVisitor import ToCairoVisitor
from yul.utils import (
    get_for_contract,
    get_function_mutabilities,
    get_public_functions,
    make_abi_StarkNet_encodable,
)
from yul.WarpException import warp_assert

AST_GENERATOR = "kudu"


def transpile_from_solidity(sol_src_path, main_contract) -> dict:
    sol_src_path_modified = sol_src_path[:-4] + "_marked.sol"
    if not shutil.which(AST_GENERATOR):
        sys.exit(f"Please install {AST_GENERATOR} first")

    try:
        result = subprocess.run(
            [AST_GENERATOR, sol_src_path, main_contract],
            check=True,
            capture_output=True,
        )
    except subprocess.CalledProcessError as e:
        print(e.stderr.decode("utf-8"), file=sys.stderr)
        raise e
    with open(sol_src_path_modified) as f:
        sol_source = f.read()
        public_functions = get_public_functions(sol_source)
        function_mutabilities = get_function_mutabilities(sol_source)
        output = get_for_contract(sol_source, main_contract, ["abi", "bin"])
        warp_assert(
            output,
            f"Couldn't extract {main_contract}'s abi and bytecode from {sol_src_path}",
        )
        abi, bytecode = output
    yul_ast = parse_node(json.loads(result.stdout))
    cairo_code, dynamic_argument_functions = transpile_from_yul(
        yul_ast, public_functions, function_mutabilities
    )
    os.remove(sol_src_path_modified)
    return {
        "cairo_code": cairo_code,
        "sol_source": sol_source,
        "sol_abi": make_abi_StarkNet_encodable(abi),
        "sol_abi_original": abi,
        "sol_bytecode": bytecode,
        "dynamic_argument_functions": dynamic_argument_functions,
    }


def transpile_from_yul(
    yul_ast: ast.Node,
    public_functions: list[str],
    function_mutabilities: dict[str, str],
) -> str:
    name_gen = NameGenerator()
    cairo_functions = CairoFunctions(FunctionGenerator())
    yul_ast = ForLoopSimplifier().map(yul_ast)
    yul_ast = ForLoopEliminator(name_gen).map(yul_ast)
    yul_ast = MangleNamesVisitor().map(yul_ast)
    yul_ast = SwitchToIfVisitor().map(yul_ast)
    yul_ast = ExpressionSplitter(name_gen).map(yul_ast)
    yul_ast = RevertNormalizer().map(yul_ast)
    yul_ast = ScopeFlattener(name_gen).map(yul_ast)
    yul_ast = LeaveNormalizer().map(yul_ast)
    yul_ast = RevertNormalizer().map(yul_ast)
    yul_ast = FunctionPruner(public_functions).map(yul_ast)

    cairo_visitor = ToCairoVisitor(
        public_functions,
        function_mutabilities,
        name_gen,
        cairo_functions,
        get_default_builtins,
    )
    return (
        parse_file(cairo_visitor.translate(yul_ast)).format(),
        cairo_visitor.dynamic_argument_functions,
    )


def main(argv):
    if len(argv) != 3:
        sys.exit("Supply SOLIDITY-CONTRACT and MAIN-CONTRACT-NAME")
    sol_src_path = argv[1]
    main_contract = argv[2]
    transpile_from_solidity(sol_src_path, main_contract)


if __name__ == "__main__":
    main(sys.argv)



from __future__ import annotations

from typing import Optional

import yul.yul_ast as ast
from yul.AstMapper import AstMapper


class SwitchToIfVisitor(AstMapper):
    def visit_switch(self, node: ast.Switch) -> ast.Block:
        return self.visit(
            ast.Block(
                (
                    ast.VariableDeclaration(
                        variables=[ast.TypedName("match_var")], value=node.expression
                    ),
                    switch_to_if(ast.Identifier("match_var"), node.cases),
                )
            )
        )


def switch_to_if(switch_var: ast.Identifier, cases: list[ast.Case]) -> ast.Block:
    res = switch_to_if_helper(switch_var, cases, case_no=0)
    assert res is not None
    return res


def switch_to_if_helper(
    switch_var: ast.Identifier, cases: list[ast.Case], case_no: int = 0
) -> Optional[ast.Block]:
    assert case_no <= len(cases)
    if case_no == len(cases):
        return None
    if cases[case_no].value is None:
        assert case_no == len(cases) - 1, "Default case should be the last one"
        return cases[case_no].body
    return ast.Block(
        (
            ast.If(
                condition=ast.FunctionCall(
                    function_name=ast.Identifier("eq"),
                    arguments=[switch_var, cases[case_no].value],
                ),
                body=cases[case_no].body,
                else_body=switch_to_if_helper(switch_var, cases, case_no + 1),
            ),
        )
    )



from __future__ import annotations

from contextlib import contextmanager
from dataclasses import dataclass
from typing import Optional

import yul.yul_ast as ast
from yul.AstMapper import AstMapper
from yul.NameGenerator import NameGenerator


@dataclass
class BlockEnv:
    def __init__(self, name_gen: NameGenerator):
        self.name_gen = name_gen
        self.split_stmts: list[ast.VariableDeclaration] = []

    def process_subexpr(self, call: ast.FunctionCall) -> str:
        name = self.name_gen.make_subexpr_name()
        self.split_stmts.append(
            ast.VariableDeclaration(variables=[ast.TypedName(name)], value=call)
        )
        return name


class ExpressionSplitter(AstMapper):
    def __init__(self, name_gen: NameGenerator):
        super().__init__()
        self.name_gen = name_gen
        self.env: Optional[BlockEnv] = None

    def visit_function_call(self, node: ast.FunctionCall):
        if len(self.path) == 1:  # no parent
            return ast.FunctionCall(node.function_name, self.visit_list(node.arguments))
        parent = self.path[-2]
        if isinstance(
            parent, (ast.Assignment, ast.VariableDeclaration, ast.ExpressionStatement)
        ):
            return ast.FunctionCall(node.function_name, self.visit_list(node.arguments))
        var_name = self.env.process_subexpr(node)
        return ast.Identifier(var_name)

    def visit_block(self, node: ast.Block):
        with self._new_block():
            new_stmts = []
            for stmt in node.statements:
                new_stmt = self.visit(stmt)
                split_stmts = [new_stmt]  # in the reverse order of declaration
                while self.env.split_stmts:
                    new_split_stmts = self.env.split_stmts
                    self.env.split_stmts = []
                    split_stmts.extend(self.visit_list(new_split_stmts))
                split_stmts.reverse()
                new_stmts.extend(split_stmts)
            return ast.Block(tuple(new_stmts))

    @contextmanager
    def _new_block(self) -> BlockEnv:
        with self.name_gen.new_block():
            old_env = self.env
            self.env = BlockEnv(self.name_gen)
            try:
                yield None
            finally:
                self.env = old_env



from __future__ import annotations

import re
from dataclasses import dataclass
from typing import Iterable, Optional

from yul.WarpException import warp_assert

GETTER_PATTERN = re.compile(r"getter_fun_(\S*)(_(\d+))?")
SETTER_PATTERN = re.compile(r"setter_fun_(\S*)(_(\d+))?")


def extract_var_from_getter(getter: str) -> Optional[str]:
    match = re.fullmatch(GETTER_PATTERN, getter)
    if not match:
        return None
    return match.group(1)


def extract_var_from_setter(setter: str) -> Optional[str]:
    match = re.fullmatch(SETTER_PATTERN, setter)
    if not match:
        return None
    return match.group(1)


@dataclass(frozen=True, order=True)
class StorageVar:
    name: str
    arg_types: tuple[str]
    res_type: str


def generate_getter_body(getter_var: str, args: Iterable[str]) -> str:
    args_repr = ", ".join(f"{x}.low, {x}.high" for x in args)
    return f"let (res) = {getter_var}.read({args_repr})\nreturn (res)"


def generate_setter_body(setter_var: str, args: Iterable[str]) -> str:
    *access_args, value_arg = args
    args_repr = ", ".join(f"{x}.low, {x}.high" for x in access_args)
    if access_args:
        args_repr += ", "
    return f"{setter_var}.write({args_repr}{value_arg})\nreturn ()"


def generate_storage_var_declaration(var: StorageVar) -> str:
    warp_assert(
        all(x == "Uint256" for x in var.arg_types),
        "We don't support storage variables parameterized by types other than Uint256",
    )
    warp_assert(
        var.res_type == "Uint256",
        "We don't support storage variables that return types other than Uint256",
    )
    args_repr = ", ".join(
        f"arg{i}_low, arg{i}_high" for i, typ in enumerate(var.arg_types)
    )
    return (
        f"@storage_var\n"
        f"func {var.name}({args_repr}) -> (res: {var.res_type}):\n"
        f"end\n"
    )



from __future__ import annotations

import abc
import sys
from abc import ABC
from typing import Callable, Mapping, Optional, Sequence

from yul.FunctionGenerator import CairoFunctions, FunctionInfo
from yul.Imports import Imports
from yul.utils import get_low_bits

UINT256_MODULE = "starkware.cairo.common.uint256"


class BuiltinHandler(ABC):
    """Describes how an operation built into Yul translates to Cairo.

    References:
    https://docs.soliditylang.org/en/latest/yul.html?highlight=optimization#evm-dialect

    """

    def get_function_call(self, function_args: Sequence[str]) -> str:
        """Given the builtin operation and arguments, returns cairo code for
        this operation.

        - 'function_args' — a sequence of arguments already translated to Cairo

        """
        kwarg_names = self.get_kwarg_names()
        function_name = self.get_function_name()
        if kwarg_names:
            assert len(kwarg_names) == len(function_args)
            args_repr = ", ".join(
                f"{kw}={v}" for kw, v in zip(kwarg_names, function_args)
            )
        else:
            args_repr = ", ".join(function_args)
        return f"{function_name}({args_repr})"

    def required_imports(self) -> Imports:
        """Specifies Cairo imports requried for this operation."""
        module = self.get_module()
        if module is None:
            return {}
        return {module: {self.get_function_name()}}

    @abc.abstractmethod
    def get_function_name(self) -> str:
        """Specifies a name for the Cairo function used for this builtin."""
        pass

    @abc.abstractmethod
    def get_module(self) -> Optional[str]:
        """Specifies a module which needs to be imported. 'None' if no module
        is needed.

        """
        pass

    @abc.abstractmethod
    def get_used_implicits(self) -> tuple[str, ...]:
        """Specifies implicits needed for the Cairo function"""
        pass

    @abc.abstractmethod
    def get_kwarg_names(self) -> Optional[tuple[str, ...]]:
        """Specifies names of arguments of the Cairo function.

        If not 'None', all passed arguments will be passed as keyword
        arguments, i.e. 'ARG_NAME=' will be prepended to them.

        Right now, it's either all arguments are keyword arguments or
        none are. In the future, we can allow passing some names as
        'None'. That would mean that this arguments don't need the
        argument name prepending.

        """
        pass


class StaticHandler(BuiltinHandler):
    """A 'BuiltinHandler' for which function parameters are known at the
    handler creation time.

    """

    def __init__(
        self,
        function_name: str,
        module: Optional[str] = None,
        used_implicits: tuple[str, ...] = ("range_check_ptr",),
        kwarg_names: Optional[tuple[str, ...]] = None,
    ):
        self.function_name = function_name
        self.module = module
        self.used_implicits = used_implicits
        self.kwarg_names = kwarg_names

    def get_function_name(self):
        return self.function_name

    def get_module(self):
        return self.module

    def get_used_implicits(self):
        return self.used_implicits

    def get_kwarg_names(self):
        return self.kwarg_names


class DynamicHandler(BuiltinHandler):
    """A 'BuiltinHandler' for functions that are generated at the call
    time.

    """

    def __init__(self, info_gen: Callable[[], FunctionInfo]):
        """'info_gen' is a function that generates the required cairo function
        and returns information about it. Amortized it should run
        fast. See 'yul.FunctionGenerator'.

        """
        self.info_gen = info_gen

    def get_function_name(self):
        return self.info_gen().name

    def get_module(self):
        return None

    def get_used_implicits(self):
        return self.info_gen().implicits

    def get_kwarg_names(self):
        return self.info_gen().kwarg_names


class NotImplementedOp(DynamicHandler):
    def __init__(self, cairo_functions: CairoFunctions):
        super().__init__(cairo_functions.stubbing_function)

    def get_function_call(self, _args: Sequence[str]):
        print(
            f"WARNING: This contract referenced '{type(self)}' "
            f"which is not yet supported. Evaluating"
            f"this contract on starknet may result in unexpected"
            f"behavior.",
            file=sys.stderr,
        )
        return super().get_function_call([])


# ============ Comparisons ============
class IsZero(StaticHandler):
    def __init__(self):
        super().__init__(function_name="is_zero", module="evm.uint256")


class Eq(StaticHandler):
    def __init__(self):
        super().__init__(function_name="is_eq", module="evm.uint256")


class Lt(StaticHandler):
    def __init__(self):
        super().__init__(function_name="is_lt", module="evm.uint256")


class Gt(StaticHandler):
    def __init__(self):
        super().__init__(function_name="is_gt", module="evm.uint256")


class Slt(StaticHandler):
    def __init__(self):
        super().__init__(function_name="slt", module="evm.uint256")


class Sgt(StaticHandler):
    def __init__(self):
        super().__init__(function_name="sgt", module="evm.uint256")


# ============ Bitwise ============
class And(StaticHandler):
    def __init__(self):
        super().__init__(function_name="uint256_and", module=UINT256_MODULE)


class Or(StaticHandler):
    def __init__(self):
        super().__init__(function_name="uint256_or", module=UINT256_MODULE)


class Not(StaticHandler):
    def __init__(self):
        super().__init__(function_name="uint256_not", module=UINT256_MODULE)


class Xor(StaticHandler):
    def __init__(self):
        super().__init__(function_name="uint256_xor", module=UINT256_MODULE)


class Shl(StaticHandler):
    def __init__(self):
        super().__init__(function_name="u256_shl", module="evm.uint256")


class Shr(StaticHandler):
    def __init__(self):
        super().__init__(function_name="u256_shr", module="evm.uint256")


class Sar(StaticHandler):
    def __init__(self):
        super().__init__(function_name="uint256_sar", module="evm.uint256")


class Byte(StaticHandler):
    def __init__(self):
        super().__init__(function_name="uint256_byte", module="evm.uint256")


# ============ Arithmetic ============
class Add(StaticHandler):
    def __init__(self):
        super().__init__(function_name="u256_add", module="evm.uint256")


class Mul(StaticHandler):
    def __init__(self):
        super().__init__(function_name="u256_mul", module="evm.uint256")


class Sub(StaticHandler):
    def __init__(self):
        super().__init__(function_name="uint256_sub", module=UINT256_MODULE)


class Div(StaticHandler):
    def __init__(self):
        super().__init__(function_name="u256_div", module="evm.uint256")


class Sdiv(StaticHandler):
    def __init__(self):
        super().__init__(function_name="u256_sdiv", module="evm.uint256")


class Exp(StaticHandler):
    def __init__(self):
        super().__init__(function_name="uin256_exp", module=UINT256_MODULE)


class Mod(StaticHandler):
    def __init__(self):
        super().__init__(function_name="uint256_mod", module="evm.uint256")


class SMod(StaticHandler):
    def __init__(self):
        super().__init__(function_name="smod", module="evm.uint256")


class AddMod(StaticHandler):
    def __init__(self):
        super().__init__(function_name="uin256_addmod", module="evm.uint256")


class MulMod(StaticHandler):
    def __init__(self):
        super().__init__(function_name="uin256_mulmod", module="evm.uint256")


class SignExtend(StaticHandler):
    def __init__(self):
        super().__init__(function_name="uin256_signextend", module="evm.uint256")


# ============ Memory ============
class MStore(StaticHandler):
    def __init__(self):
        super().__init__(
            function_name="mstore_",
            module="evm.memory",
            used_implicits=("memory_dict", "msize", "range_check_ptr"),
            kwarg_names=("offset", "value"),
        )

    def get_function_call(self, args: Sequence[str]):
        (address, value) = args
        return super().get_function_call([get_low_bits(address), value])


class MStore8(StaticHandler):
    def __init__(self):
        super().__init__(
            function_name="mstore8_",
            module="evm.memory",
            used_implicits=("memory_dict", "msize", "range_check_ptr"),
        )

    def get_function_call(self, args: Sequence[str]):
        (address, value) = args
        return super().get_function_call([get_low_bits(address), value])


class MLoad(StaticHandler):
    def __init__(self):
        super().__init__(
            function_name="mload_",
            module="evm.memory",
            used_implicits=("memory_dict", "msize", "range_check_ptr"),
        )

    def get_function_call(self, args: Sequence[str]):
        (address,) = args
        return super().get_function_call([get_low_bits(address)])


class MSize(StaticHandler):
    def __init__(self):
        super().__init__(
            function_name="get_msize", module="evm.memory", used_implicits=("msize",)
        )


# ============ Storage ============
class SStore(DynamicHandler):
    def __init__(self, cairo_functions: CairoFunctions):
        super().__init__(cairo_functions.sstore_function)


class SLoad(DynamicHandler):
    def __init__(self, cairo_functions: CairoFunctions):
        super().__init__(cairo_functions.sload_function)

    def get_kwarg_names(self):
        return None


# ============ Keccak ============
class SHA3(StaticHandler):
    def __init__(self):
        super().__init__(
            function_name="sha",
            module="evm.sha3",
            used_implicits=("memory_dict", "msize", "range_check_ptr"),
        )

    def get_function_call(self, args: Sequence[str]):
        (offset, length) = map(get_low_bits, args)
        return super().get_function_call([offset, length])


# ============ Call Data ============
class Caller(StaticHandler):
    def __init__(self):
        super().__init__(
            function_name="get_caller_data_uint256",
            module="evm.calls",
            used_implicits=("syscall_ptr",),
        )


class CallDataLoad(StaticHandler):
    def __init__(self):
        super().__init__(
            function_name="calldata_load",
            module="evm.calls",
            used_implicits=("range_check_ptr", "exec_env"),
        )

    def get_function_call(self, args: Sequence[str]):
        (offset,) = args
        return super().get_function_call([get_low_bits(offset)])


class CallDataSize(StaticHandler):
    def __init__(self):
        super().__init__(
            function_name="calldatasize_",
            module="evm.calls",
            used_implicits=("range_check_ptr", "exec_env"),
        )


class CallDataCopy(StaticHandler):
    def __init__(self):
        super().__init__(
            function_name="calldatacopy_",
            module="evm.calls",
            used_implicits=("range_check_ptr", "exec_env", "memory_dict", "msize"),
        )


# ============ Return Data ============


class ReturnDataCopy(StaticHandler):
    def __init__(self):
        super().__init__(
            function_name="returndata_copy",
            module="evm.calls",
            used_implicits=("range_check_ptr", "exec_env", "memory_dict"),
        )


class ReturnDataSize(DynamicHandler):
    def __init__(self, cairo_functions: CairoFunctions):
        super().__init__(lambda: cairo_functions.returndata_size_function())


class Return(StaticHandler):
    def __init__(self):
        super().__init__(
            function_name="returndata_write",
            module="evm.calls",
            used_implicits=("exec_env",),
        )


class Address(DynamicHandler):
    def __init__(self, cairo_functions: CairoFunctions):
        super().__init__(cairo_functions.address_function)


class Gas(DynamicHandler):
    def __init__(self, cairo_functions: CairoFunctions):
        super().__init__(lambda: cairo_functions.constant_function(10 ** 40))


class Delegatecall(NotImplementedOp):
    pass


class Call(StaticHandler):
    def __init__(self):
        super().__init__(
            function_name="warp_call",
            module="evm.calls",
            used_implicits=(
                "syscall_ptr",
                "exec_env",
                "memory_dict",
                "range_check_ptr",
                "bitwise_ptr",
            ),
        )


class StaticCall(StaticHandler):
    def __init__(self):
        super().__init__(
            function_name="warp_static_call",
            module="evm.calls",
            used_implicits=(
                "syscall_ptr",
                "exec_env",
                "memory_dict",
                "range_check_ptr",
                "bitwise_ptr",
            ),
        )


class ExtCodeSize(DynamicHandler):
    def __init__(self, cairo_functions: CairoFunctions):
        super().__init__(lambda: cairo_functions.constant_function(1))


class CallValue(DynamicHandler):
    def __init__(self, cairo_functions: CairoFunctions):
        super().__init__(lambda: cairo_functions.constant_function(0))


def get_default_builtins(
    cairo_functions: CairoFunctions,
) -> Mapping[str, BuiltinHandler]:
    """Returns a mapping from default Yul builtins to their handlers."""
    return {
        "add": Add(),
        "addmod": AddMod(),
        "address": Address(cairo_functions),
        "and": And(),
        "byte": Byte(),
        "call": Call(),
        "calldatacopy": CallDataCopy(),
        "calldataload": CallDataLoad(),
        "calldatasize": CallDataSize(),
        "caller": Caller(),
        "callvalue": CallValue(cairo_functions),
        "delegatecall": Delegatecall(cairo_functions),
        "div": Div(),
        "eq": Eq(),
        "exp": Exp(),
        "extecodesize": ExtCodeSize(cairo_functions),
        "gas": Gas(cairo_functions),
        "gt": Gt(),
        "iszero": IsZero(),
        "keccak256": SHA3(),
        "lt": Lt(),
        "mload": MLoad(),
        "mod": Mod(),
        "msize": MSize(),
        "mstore": MStore(),
        "mstore8": MStore8(),
        "mul": Mul(),
        "mulmod": MulMod(),
        "not": Not(),
        "or": Sub(),
        "return": Return(),
        "returndatacopy": ReturnDataCopy(),
        "returndatasize": ReturnDataSize(cairo_functions),
        "sar": Sar(),
        "sdiv": Sdiv(),
        "sgt": Sgt(),
        "shl": Shl(),
        "shr": Shr(),
        "signextend": SignExtend(),
        "sload": SLoad(cairo_functions),
        "slt": Slt(),
        "smod": SMod(),
        "sstore": SStore(cairo_functions),
        "staticcall": StaticCall(),
        "sub": Sub(),
        "xor": Xor(),
    }



from __future__ import annotations

from typing import Optional

import yul.yul_ast as ast
from yul.AstMapper import AstMapper
from yul.storage_access import extract_var_from_getter, extract_var_from_setter
from yul.top_sort import top_sort_ast

REVERT = ast.FunctionCall(
    function_name=ast.Identifier("revert"), arguments=[ast.Literal(0), ast.Literal(0)]
)
REVERT_STMT = ast.ExpressionStatement(REVERT)


class RevertNormalizer(AstMapper):
    def __init__(self):
        super().__init__()
        self.revert_functions: set[str] = set()

    def map(self, node: ast.Node, **kwargs) -> ast.Node:
        if isinstance(node, ast.Block):
            node = top_sort_ast(node)
        return self.visit(node)

    def visit_assignment(self, node: ast.Assignment) -> ast.Statement:
        value = self.visit(node.value)
        if self._is_revert(value):
            return REVERT_STMT
        else:
            return ast.Assignment(variable_names=node.variable_names, value=node.value)

    def visit_function_call(self, node: ast.FunctionCall) -> ast.FunctionCall:
        if node.function_name.name in self.revert_functions:
            return REVERT
        else:
            return node

    def visit_variable_declaration(
        self, node: ast.VariableDeclaration
    ) -> ast.Statement:
        value = self.visit(node.value) if node.value else None
        if self._is_revert(value):
            return REVERT_STMT
        else:
            return ast.VariableDeclaration(variables=node.variables, value=value)

    def visit_block(self, node: ast.Block) -> ast.Block:
        stmts = self.visit_list(node.statements)
        if any(self._is_revert(x) for x in stmts):
            return ast.Block((REVERT_STMT,))
        else:
            return ast.Block(tuple(x for x in stmts if not self._is_revert_function(x)))

    def visit_function_definition(
        self, node: ast.FunctionDefinition
    ) -> ast.FunctionDefinition:
        body = self.visit(node.body)
        if not _is_untouchable(node) and self._is_revert(body):
            self.revert_functions.add(node.name)
        return ast.FunctionDefinition(
            name=node.name,
            parameters=node.parameters,
            return_variables=node.return_variables,
            body=body,
        )

    def visit_if(self, node: ast.If) -> ast.Statement:
        condition = self.visit(node.condition)
        if self._is_revert(condition):
            return ast.ExpressionStatement(REVERT)
        body = self.visit(node.body)
        else_body = self.visit(node.else_body) if node.else_body else None
        if self._is_revert(body) and self._is_revert(else_body):
            return REVERT_STMT
        return ast.If(condition, body, else_body)

    def _is_revert(self, node: Optional[ast.Node]):
        if isinstance(node, ast.FunctionCall):
            return node.function_name.name == "revert"
        if isinstance(node, ast.ExpressionStatement):
            return self._is_revert(node.expression)
        if isinstance(node, ast.Block):
            return len(node.statements) == 1 and self._is_revert(node.statements[0])
        return False

    def _is_revert_function(self, node: ast.Node):
        return (
            isinstance(node, ast.FunctionDefinition)
            and node.name in self.revert_functions
        )


def _is_untouchable(function: ast.FunctionDefinition) -> bool:
    name = function.name
    if name == "fun_ENTRY_POINT":
        return True
    if extract_var_from_getter(name) or extract_var_from_setter(name):
        return True
    return False



import yul.yul_ast as ast
from yul.AstMapper import AstMapper


class ForLoopSimplifier(AstMapper):
    def visit_for_loop(self, node: ast.ForLoop):
        return ast.Block(
            self.visit(node.pre).statements
            + (
                ast.ForLoop(
                    pre=ast.Block(),
                    condition=node.condition,
                    post=ast.Block(),
                    body=self.visit(
                        ast.Block(node.body.statements + node.post.statements)
                    ),
                ),
            )
        )



from __future__ import annotations

import yul.yul_ast as ast
from yul.AstMapper import AstMapper
from yul.call_graph import CallGraph, build_callgraph


class FunctionPruner(AstMapper):
    def __init__(self, public_functions: list[str]):
        super().__init__()
        self.public_functions: frozenset[str] = frozenset(public_functions)
        self.callgraph: CallGraph = {}
        self.visited_functions: set[str] = set()

    def map(self, node: ast.Node, *args, **kwargs) -> ast.Node:
        if not isinstance(node, ast.Block):
            return self.visit(node, *args, **kwargs)

        self.callgraph = build_callgraph(node)
        for function in self.callgraph:
            f_name = function.name
            if (
                f_name in self.public_functions
                or f_name == "fun_ENTRY_POINT"
                or "warp_constructor" in f_name
            ):
                self._dfs(function)

        return self.visit(node, *args, **kwargs)

    def visit_block(self, node: ast.Block):
        statements = []
        for stmt in node.statements:
            if self._is_unused_function(stmt):
                continue
            statements.append(self.visit(stmt))

        return ast.Block(statements=tuple(statements))

    def _dfs(self, function: ast.FunctionDefinition):
        if function.name in self.visited_functions:
            return

        self.visited_functions.add(function.name)
        for f in self.callgraph[function]:
            self._dfs(f)

    def _is_unused_function(self, node):
        return (
            isinstance(node, ast.FunctionDefinition)
            and node.name not in self.visited_functions
            and "constructor" not in node.name
            and "setter" not in node.name
            and "getter" not in node.name
        )



import os
import pathlib
from io import open
from os import path, environ
import shutil

from setuptools import find_packages, setup
import pkg_resources

# The directory containing this file
HERE = pathlib.Path(__file__).parent
# The text of the README file
with open(path.join(HERE, "README.md")) as f:
    README = f.read()

# automatically captured required modules for install_requires in requirements.txt and as well as configure dependency links
with open(path.join(HERE, "requirements.txt"), encoding="utf-8") as f:
    all_reqs = f.read().split("\n")
install_requires = [
    x.strip()
    for x in all_reqs
    if ("git+" not in x) and (not x.startswith("#")) and (not x.startswith("-"))
]
dependency_links = [x.strip().replace("git+", "") for x in all_reqs if "git+" not in x]

setup(
    name="evm-warp",
    description="Transpile EVM-Compatible Languages To Cairo",
    version="0.2.1",
    package_dir={"": "warp"},
    packages=["cairo-src", "cairo-src.evm", "cli", "yul"],  # list of all packages
    include_package_data=True,
    package_data={"": ["*.json", "*.cairo"]},
    install_requires=install_requires,
    python_requires=">=3.7",  # any python greater than 3.7
    entry_points="""
        [console_scripts]
        warp=cli.warp_cli:main
    """,
    author="Nethermind",
    keyword="Ethereum, Layer2, ETH, StarkNet, Nethermind, StarkWare, transpilation, warp, transpiler, cairo",
    long_description=README,
    long_description_content_type="text/markdown",
    license="Apache 2.0",
    url="https://github.com/NethermindEth/warp",
    download_url="",
    dependency_links=dependency_links,
    author_email="hello@nethermind.io",
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: Apache Software License",
    ],
)

# try:
# base_env_dir = environ['VIRTUAL_ENV']
# old_kudu_exe = path.join(pkg_resources.get_distribution("warp==0.1.0").location, "bin/kudu")
# new_kudu_exe = path.join(base_env_dir,"bin/kudu")
# print(new_kudu_exe)
# print(old_kudu_exe)
# if os.path.exists(new_kudu_exe):
#     os.remove(new_kudu_exe)
# shutil.move(old_kudu_exe, new_kudu_exe)
# print(pkg_resources.get_distribution("warp==0.1.0").location)
# except KeyError:
#     base_


from __future__ import annotations

import json
import re
from typing import List, Optional, Sequence

import solcx

UPPERCASE_PATTERN = re.compile(r"[A-Z]")

STATEMENT_STRINGS = {
    "ExpressionStatement",
    "Assignment",
    "VariableDeclaration",
    "FunctionDefinition",
    "If",
    "Switch",
    "ForLoop",
    "Break",
    "Continue",
    "Leave",
    "Block",
}

UNSUPPORTED_BUILTINS = [
    "balance",
    "basefee",
    "blockhash",
    "chainid",
    "codecopy",
    "codesize",
    "coinbase",
    "create2",
    "delegatecall",
    "difficulty",
    "extcodecopy",
    "extcodehash",
    "extcodesize",
    "gaslimit",
    "gasprice",
    "invalid",
    "log0",
    "log1",
    "log2",
    "log3",
    "log4",
    "number",
    "origin",
    "pc",
    "selfbalance",
    "selfdestruct",
    "timestamp",
]

HANDLERS_DECL = """
@storage_var
func this_address() -> (res: felt):
end

@storage_var
func address_initialized() -> (res : felt):
end

func initialize_address{range_check_ptr, syscall_ptr: felt*, pedersen_ptr : HashBuiltin*}(self_address : felt):
    let (address_init) = address_initialized.read()
    if address_init == 1:
        return ()
    end
    this_address.write(self_address)
    address_initialized.write(1)
    return ()
end
"""


def get_low_bits(string: str) -> str:
    return get_low_high(string)[0]


def get_low_high(string: str) -> str:
    try:
        value = int(string)
        high, low = divmod(value, 2 ** 128)
        return f"{low}", f"{high}"
    except ValueError:
        return f"{string}.low", f"{string}.high"


def is_statement(node):
    return type(node).__name__ in STATEMENT_STRINGS


def remove_prefix(text, prefix):
    if text.startswith(prefix):
        return text[len(prefix) :]
    return text


def snakify(camel_case: str) -> str:
    """ThisCaseWord -> this_case_word"""
    return remove_prefix(
        UPPERCASE_PATTERN.sub(lambda m: f"_{m.group(0).lower()}", camel_case), "_"
    )


def camelize(snake_case: str) -> str:
    """this_case_word -> ThisCaseWord"""
    parts = snake_case.split("_")
    if not parts:
        return snake_case
    if any(x == "" for x in parts):
        raise ValueError(
            f"Can't camelize {snake_case}."
            " It probably contains several consecutive underscores."
        )
    return "".join(x.capitalize() for x in parts)


def get_public_functions(sol_source: str) -> list[str]:
    validate_solc_ver(sol_source)

    public_functions = set()
    abi = solcx.compile_source(sol_source, output_values=["hashes"])
    for value in abi.values():
        for v in value["hashes"]:
            public_functions.add(f"fun_{v[:v.find('(')]}")
    return list(public_functions)


def get_function_mutabilities(sol_source: str) -> dict[str, str]:
    validate_solc_ver(sol_source)

    function_visibilities = dict()
    abi = solcx.compile_source(sol_source)

    for value in abi.values():
        for v in value["abi"]:
            if v["type"] == "function":
                function_visibilities[v["name"]] = v["stateMutability"]
    return function_visibilities


def get_for_contract(
    sol_source: str, target_contract: str, output_values: list[str]
) -> Optional[Sequence[str]]:
    validate_solc_ver(sol_source)
    compiled = solcx.compile_source(sol_source, output_values=output_values)
    for contract, values in compiled.items():
        name = contract[contract.find(":") + 1 :]
        if name == target_contract:
            return [values[x] for x in output_values]
    return None


def validate_solc_ver(sol_source):
    solc_version: float = get_source_version(sol_source)
    src_ver: str = check_installed_solc(solc_version)
    solcx.set_solc_version(src_ver)


def get_source_version(sol_source: str) -> float:
    code_split = sol_source.split("\n")
    for line in code_split:
        if "pragma" in line:
            ver: float = float(line[line.index("0.") + 2 :].replace(";", ""))
            if ver < 7.6:
                raise Exception(
                    "Please use a version of solidity that is at least 0.7.6"
                )
            return ver
    raise Exception("No Solidity version specified in contract")


def check_installed_solc(source_version: float) -> str:
    solc_vers = solcx.get_installed_solc_versions()
    vers_clean = []
    src_ver = "0." + str(source_version)
    for ver in solc_vers:
        vers_clean.append(".".join(str(x) for x in list(ver.precedence_key)[:3]))
    if src_ver not in vers_clean:
        solcx.install_solc(src_ver)
    return src_ver


def cairoize_bytes(bs: bytes) -> tuple(List[int], int):
    """Represent bytes as an array of 128-bit big-endian integers and
    return a number of unused bytes in the last array cell.
    """
    unused_bytes = -len(bs) % 16
    bs = bs.ljust(len(bs) + unused_bytes, b"\x00")
    arr = [int.from_bytes(bs[i : i + 16], "big") for i in range(0, len(bs), 16)]
    return (arr, unused_bytes)


def make_abi_StarkNet_encodable(abi):
    """
    Because we need 31 byte addresses for starknet, we need to encode them as uint256,
    but if we pass an abi with address types to solcx, it will try to validate the address by
    checksum & making sure it's 20-bytes. So we just change the address abi types to uint256.
    This is fine because our fork of solc treats address types as uint256 anyway.
    We will replace this hackery with our own encoder in the future.
    """
    lol = f"{abi}"
    abiStr = lol.replace("'address'", "'uint256'")
    abiStr = abiStr.replace("'", '"')
    return json.loads(abiStr)



from __future__ import annotations

from typing import Callable

import yul.yul_ast as ast
from yul.Scope import get_scope


def extract_block_as_function(
    block: ast.Block, name: str, has_leave: bool = False
) -> tuple[ast.FunctionDefinition, ast.Statement]:
    read_vars = get_scope(block).read_variables
    if has_leave:
        # If there is a leave in the block, some subset of modified
        # variables will also be read at the time of "leaving". We
        # play safe and mark all of the modified variables as read. An
        # opportunity of optimization.
        read_vars |= get_scope(block).modified_variables
    read_vars = sorted(read_vars)
    mod_vars = sorted(get_scope(block).modified_variables)
    typed_read_vars = [ast.TypedName(x.name) for x in read_vars]
    typed_mod_vars = [ast.TypedName(x.name) for x in mod_vars]
    fun_def = ast.FunctionDefinition(
        name=name,
        parameters=typed_read_vars,
        return_variables=typed_mod_vars,
        body=block,
    )
    fun_call = ast.FunctionCall(ast.Identifier(name), read_vars)
    fun_stmt = ast.Assignment(variable_names=mod_vars, value=fun_call)
    return fun_def, fun_stmt


DUMMY_CALL = ast.Assignment([], ast.FunctionCall(ast.Identifier("__WARP_DUMMY"), []))


def extract_rec_block_as_function(
    rec_block: Callable[[ast.Statement], ast.Block], name: str, has_leave: bool = False
) -> tuple[ast.FunctionDefinition, ast.Statement]:
    stubbed_body = rec_block(DUMMY_CALL)
    read_vars = get_scope(stubbed_body).read_variables
    if has_leave:
        # If there is a leave in the block, some subset of modified
        # variables will also be read at the time of "leaving". We
        # play safe and mark all of the modified variables as read. An
        # opportunity of optimization.
        read_vars |= get_scope(stubbed_body).modified_variables
    read_vars = sorted(read_vars)
    mod_vars = sorted(get_scope(stubbed_body).modified_variables)
    typed_read_vars = [ast.TypedName(x.name) for x in read_vars]
    typed_mod_vars = [ast.TypedName(x.name) for x in mod_vars]
    fun_call = ast.FunctionCall(ast.Identifier(name), read_vars)
    fun_stmt = ast.Assignment(variable_names=mod_vars, value=fun_call)
    real_body = rec_block(fun_stmt)
    fun_def = ast.FunctionDefinition(
        name=name,
        parameters=typed_read_vars,
        return_variables=typed_mod_vars,
        body=real_body,
    )
    return fun_def, fun_stmt



from __future__ import annotations

from contextlib import contextmanager
from functools import lru_cache
from typing import Optional

import yul.yul_ast as ast
from yul.AstMapper import AstMapper
from yul.extract_block import extract_block_as_function
from yul.NameGenerator import NameGenerator
from yul.Scope import get_scope


class ScopeFlattener(AstMapper):
    def __init__(self, name_gen: NameGenerator):
        super().__init__()
        self.name_gen = name_gen
        self.block_functions: list[ast.FunctionDefinition] = []
        self.revert_created: bool = False
        self.leave_name: Optional[str] = None
        self.pass_leaves_higher = False

    def map(self, node: ast.Node, **kwargs) -> ast.Node:
        if not isinstance(node, ast.Block):
            return self.visit(node)
        all_statements = []
        statements = node.statements
        while statements:
            all_statements.extend(self.visit_list(statements))
            statements = self.block_functions
            self.block_functions = []
        return ast.Block(tuple(all_statements))

    def visit_block(self, node: ast.Block, inline: bool = False) -> ast.Block:
        if inline or not get_scope(node).bound_variables:
            stmts = []
            for stmt in node.statements:
                new_stmt = self.visit(stmt)
                if isinstance(new_stmt, ast.Block):
                    stmts.extend(new_stmt.statements)
                else:
                    stmts.append(new_stmt)
            return ast.Block(tuple(stmts))

        with self._new_extraction():
            fun_name = self.name_gen.make_block_name()
            node = super().visit_block(node)
            block_fun, block_stmt = extract_block_as_function(node, fun_name)
            self.block_functions.append(block_fun)
            leave_id = ast.Identifier(self.leave_name)
            if leave_id not in get_scope(node).modified_variables:
                stmts = (block_stmt,)
            else:
                stmts = (block_stmt, ast.If(condition=leave_id, body=ast.LEAVE_BLOCK))
            return ast.Block(stmts)

    def visit_function_definition(self, node: ast.FunctionDefinition):
        # We do not want to flatten an if-block if it is the only statement in
        # the function body. Skip visit_if in such cases and instead visit all
        # the branches of the if-block.

        with self._new_function():
            if len(node.body.statements) == 1 and isinstance(
                node.body.statements[0], ast.If
            ):
                if_stmt = node.body.statements[0]
                if_body = self.visit_block(if_stmt.body, inline=True)
                if_else_body = None
                if if_stmt.else_body:
                    if_else_body = self.visit_block(if_stmt.else_body, inline=True)
                body = ast.Block((ast.If(if_stmt.condition, if_body, if_else_body),))
            else:
                body = self.visit(node.body, inline=True)
            leave_id = ast.Identifier(self.leave_name)
            if leave_id in get_scope(body).modified_variables:
                leave_var_decl = ast.VariableDeclaration(
                    [ast.TypedName(self.leave_name)], ast.Literal(False)
                )
                body = ast.Block((leave_var_decl, *body.statements))
            uninitialized_return_vars: list[ast.TypedName] = []
            params_set: set[str] = {x.name for x in node.parameters}
            for v in node.return_variables:
                v_id = ast.Identifier(v.name)
                if v_id in get_scope(body).read_variables and v.name not in params_set:
                    uninitialized_return_vars.append(v)
            if uninitialized_return_vars:
                init = ast.VariableDeclaration(uninitialized_return_vars, value=None)
                body = ast.Block((init, *body.statements))
            return ast.FunctionDefinition(
                name=node.name,
                parameters=self.visit_list(node.parameters),
                return_variables=self.visit_list(node.return_variables),
                body=body,
            )

    def visit_if(self, node: ast.If):
        if self.is_leave_if(node):
            return super().visit_if(node)

        revert_if = self.is_revert_if(node)
        if revert_if:
            fun_name = self.name_gen.take_cond_revert_name()
        else:
            fun_name = self.name_gen.make_if_name()

        with self._new_extraction():
            if_block = ast.Block((super().visit_if(node),))
            if_fun, if_stmt = extract_block_as_function(if_block, fun_name)
            if revert_if and self.revert_created:
                return if_stmt
            self.block_functions.append(if_fun)
            self.revert_created |= revert_if
            leave_id = ast.Identifier(self.leave_name)
            if leave_id in get_scope(if_block).modified_variables:
                stmts = (if_stmt, ast.If(condition=leave_id, body=ast.LEAVE_BLOCK))
            else:
                stmts = (if_stmt,)
            return ast.Block(stmts)

    def visit_leave(self, node: ast.Leave):
        if not self.pass_leaves_higher:
            return node
        else:
            leave_var_assign = ast.Assignment(
                [ast.Identifier(self.leave_name)], ast.Literal(True)
            )
            return ast.Block((leave_var_assign, ast.LEAVE))

    def is_revert_if(self, node: ast.If):
        return _is_revert(node.body) or _is_revert(node.else_body)

    def is_leave_if(self, node: ast.If):
        return _is_leave(node.body) or _is_leave(node.else_body)

    @contextmanager
    def _new_function(self):
        old_leave_name = self.leave_name
        self.leave_name = self.name_gen.make_leave_name()
        try:
            yield None
        finally:
            self.leave_name = old_leave_name

    @contextmanager
    def _new_extraction(self):
        old_pass_leaves_higher = self.pass_leaves_higher
        self.pass_leaves_higher = True
        try:
            yield None
        finally:
            self.pass_leaves_higher = old_pass_leaves_higher


@lru_cache()
def _is_leave(node: Optional[ast.Node]) -> bool:
    if isinstance(node, ast.Block):
        return any(_is_leave(x) for x in node.statements)
    elif isinstance(node, ast.If):
        return _is_leave(node.condition) and (
            not node.else_body or _is_leave(node.else_body)
        )
    elif isinstance(node, ast.Leave):
        return True
    else:
        return False


def _is_revert(node: Optional[ast.Node]) -> bool:
    if isinstance(node, ast.FunctionCall):
        return node.function_name.name == "revert"
    if isinstance(node, ast.ExpressionStatement):
        return _is_revert(node.expression)
    if isinstance(node, ast.Block):
        return len(node.statements) == 1 and _is_revert(node.statements[0])
    return False



from __future__ import annotations

from dataclasses import dataclass
from typing import Callable, Iterable

from yul.implicits import print_implicit
from yul.storage_access import StorageVar, generate_getter_body, generate_setter_body


class FunctionGenerator:
    """Utility to create multi-use functions."""

    def __init__(self):
        self.definitions: dict[str, str] = {}

    def create_function(self, name: str, definition: Callable[[], str]) -> None:
        """Creates definition of a function named 'name', unless it already
        exists.

        'definition' is a function generator.

        """
        self.definitions.setdefault(name, definition())


@dataclass
class FunctionInfo:
    """Represents all information necessary to make a call to a particular Cairo function.

    'name' is a name of the function.
    'implicits' is a sequence of implicits needed by the functions.
    'kwarg_names' is a sequence of names of the arguments of the function.

    """

    name: str
    implicits: tuple[str, ...]
    kwarg_names: tuple[str, ...]


class CairoFunctions:
    """A collection of functions generators. All function generators can
    be used to generate a function and get all information necessary
    to make a call to it (via 'FunctionInfo'). No function is generated
    twice.

    """

    def __init__(self, generator: FunctionGenerator):
        self.generator = generator
        self.storage_vars: set[StorageVar] = set()

    def get_definitions(self) -> Iterable[str]:
        """Returns all generated function definitions. Note, it doesn't
        include generated 'StorageVar's.

        """
        yield from self.generator.definitions.values()

    def constant_function(self, constant: int) -> FunctionInfo:
        """Creates a function that always returns a specified 'constant'."""
        name = f"__warp_constant_{constant}"

        def inner():
            high, low = divmod(constant, 2 ** 128)
            return "\n".join(
                [
                    f"func {name}() -> (res: Uint256):",
                    f"return (Uint256(low={low}, high={high}))",
                    f"end\n",
                ]
            )

        self.generator.create_function(name, inner)
        return FunctionInfo(name=name, implicits=(), kwarg_names=())

    def identity_function(self, types: list[str]):
        name = f"__warp_identity_{'__'.join(types)}"
        args = ", ".join([f"arg{i} : {tp}" for (i, tp) in enumerate(types)])
        returns = ", ".join([f"arg{i}" for i in range(len(types))])

        def inner():
            return "\n".join(
                [
                    f"func {name}({args}) -> ({args}):",
                    f"return ({returns})",
                    f"end\n",
                ]
            )

        self.generator.create_function(name, inner)
        return FunctionInfo(name=name, implicits=(), kwarg_names=())

    def sload_function(self) -> FunctionInfo:
        """Creates a function that fetches a Uint256 from the emulated Ethereum
        storage, as well as a 'StorageVar' representing the said storage.

        """
        name = "sload"
        implicits = ("range_check_ptr", "pedersen_ptr", "syscall_ptr")

        def inner():
            implicits_str = (
                "{" + ", ".join(print_implicit(x) for x in sorted(implicits)) + "}"
            )
            return "\n".join(
                [
                    f"func {name}{implicits_str}(key: Uint256) -> (value: Uint256):",
                    generate_getter_body("evm_storage", ("key",)),
                    "end\n",
                ]
            )

        self.generator.create_function(name, inner)
        self.storage_vars.add(
            StorageVar(name="evm_storage", arg_types=("Uint256",), res_type="Uint256")
        )
        return FunctionInfo(name=name, implicits=implicits, kwarg_names=("key",))

    def sstore_function(self) -> FunctionInfo:
        """Creates a function that stores a Uint256 from the emulated Ethereum
        storage, as well as a 'StorageVar' representing the said storage.

        """
        name = "sstore"
        implicits = ("range_check_ptr", "pedersen_ptr", "syscall_ptr")

        def inner():
            implicits_str = (
                "{" + ", ".join(print_implicit(x) for x in sorted(implicits)) + "}"
            )
            return "\n".join(
                [
                    f"func {name}{implicits_str}(key: Uint256, value: Uint256):",
                    generate_setter_body("evm_storage", ("key", "value")),
                    "end\n",
                ]
            )

        self.generator.create_function(name, inner)
        self.storage_vars.add(
            StorageVar(name="evm_storage", arg_types=("Uint256",), res_type="Uint256")
        )
        return FunctionInfo(
            name=name, implicits=implicits, kwarg_names=("key", "value")
        )

    def stubbing_function(self) -> FunctionInfo:
        """Creates a function that takes no arguments and returns a single
        value. The function doesn't actually compute anything but just
        reverts.

        """
        name = f"__warp_stub"

        def inner():
            return "\n".join(
                [
                    f"func {name}() -> (res: Uint256):",
                    f"assert 1 = 0",
                    f"jmp rel 0",
                    f"end\n",
                ]
            )

        self.generator.create_function(name, inner)
        return FunctionInfo(name=name, implicits=(), kwarg_names=())

    def address_function(self) -> FunctionInfo:
        """Creates a function that returns the address of this contract on
        StarkNet.

        """
        name = f"address"
        implicits = ("syscall_ptr", "range_check_ptr", "pedersen_ptr")

        def inner():
            implicits_str = (
                "{" + ", ".join(print_implicit(x) for x in sorted(implicits)) + "}"
            )
            return "\n".join(
                [
                    f"func {name}{implicits_str}() -> (res: Uint256):",
                    "let (addr) = this_address.read()",
                    "return (Uint256(low=addr, high=0))",
                    "end\n",
                ]
            )

        self.generator.create_function(name, inner)
        return FunctionInfo(name=name, implicits=implicits, kwarg_names=())

    def returndata_size_function(self) -> FunctionInfo:
        name = f"returndata_size"
        implicits = ("exec_env",)

        def inner():
            implicits_str = (
                "{" + ", ".join(print_implicit(x) for x in sorted(implicits)) + "}"
            )
            return "\n".join(
                [
                    f"func {name}{implicits_str}() -> (res: Uint256):",
                    "return (Uint256(low=exec_env.returndata_size, high=0))",
                    "end\n",
                ]
            )

        self.generator.create_function(name, inner)
        return FunctionInfo(name=name, implicits=implicits, kwarg_names=())
